// T4.C - all routines and data for printing to disk a full calendar

/*********************************************************************************
License for TCB - Bhutan Calendar software for "grub rtsis"

Copyright (c) 2009-2011 Edward Henning

Permission is hereby granted, free of charge, to any person  obtaining a copy of
this software and associated documentation files (the "Software"), to deal in the
Software without restriction, including without limitation the rights to use, 
copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the 
Software, and to permit persons to whom the Software is furnished to do so, subject 
to the following conditions: 

The above copyright notice and this permission notice shall be included in all copies
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
************************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <conio.h>
#include "tc.h"
#include "tc.ext"
#include "bcd.h"
#include "bcd.ext"

struct ss {
  long tse[6];
  long nyi[6];
  } s[7];

void set_lang ( void );
int sol_found = 0;
long int sol_doweek;
long long int sol1, sol2, sol3; // Solstice time values.    

// Routine to check for special days, festivals, anniversaries, etc.
// Many are currently commented out. Uncomment if needed.
int chk_spec ( long int m, long int t )
  {
    switch ( m )
      {
        case 1:
          if ( t == 1L )
            {
              sprintf ( outbuf, "%s",
                "Holiday: New Year, 2 days." );
              return (1);
            }
        break;
        case 3:
          if ( t == 10L )
            {
              // Need to determine first applicable year ??
              if ( ty < 2010 || ty > 2020 )
                return (0);
              sprintf ( outbuf, "%s", "Holiday: Death Anniversary of Zhabdrung." );
              return (1);
            }
          if ( t == 15L )
            {
              sprintf ( outbuf, "%s", "Revelation of the Kalacakra Tantra." );
              return (1);
            }
        break;
        case 4:
          if ( t == 15L )
            {
              sprintf ( outbuf, "%s",
                "Holiday: Lord Buddha's Parinirvana." );
              return (1);
            }
        break;
        case 5:
          if ( t == 10L )
            {
              sprintf ( outbuf, "%s", "Holiday: Birth Anniversary of Guru Rinpoche." );
              return (1);
            }
        case 6:
          if ( t == 4L )
            {
              sprintf ( outbuf, "%s", "Holiday: The First Sermon of Lord Buddha." );
              return (1);
            }
          else if ( t == 15L )
            {
              sprintf ( outbuf, "%s",
                        "The Buddha's entry into the womb of his mother." );
              return (1);
            }
        break;
        case 8:
          if ( t == 6L )
            {
              sprintf ( outbuf, "%s",
                "Holiday: Thimpu Drubchen (Thimpu only)." );
              return (1);
            }
          else if ( t == 10L )
            {
              sprintf ( outbuf, "%s",
                "Holiday: Thimphu Tshechu (Thimpu only), three days." );
              return (1);
            }
        break;
        case 9:
          if ( t == 22L )
            {
              sprintf ( outbuf, "%s",
                "Holiday: Descending Day of Lord Buddha." );
              return (1);
            }
        break;
        case 12:
          if ( t == 1L )
            {
              sprintf ( outbuf, "%s",
                "Holiday: Traditional Day of Offering." );
              return (1);
            }
        break;
        default:
          ;
        break;
      }
    return (0);
  } // END - chk_spec ()

// Routine to check for special days, festivals, anniversaries, etc.
// As above, but this uses Gregorian dates.
int chk_west ( long int m, long int d )
  {
    switch ( m )
      {
        case 2:
          if ( d == 21L )
            {
              if ( ty < 2010 || ty > 2020 )
                return (0);            
              sprintf ( outbuf, "%s",
                "Holiday: Birth Anniversary of His Majesty the King, 3 days." );
              return (1);
            }
        break;
        case 5:
          if ( d == 2L )
            {
              if ( ty < 2010 || ty > 2020 )
                return (0);
              sprintf ( outbuf, "%s", "Holiday: Birth Anniversary of the 3rd Druk Gyalpo" );
              return (1);
            }
        break;
        case 11:
          if ( d == 11L )
            {
              if ( ty < 2010 || ty > 2020 )
                return (0);            
              sprintf ( outbuf, "%s",
                "Holiday: Birth Anniversary of the 4th Druk Gyalpo and Constitution Day." );
              return (1);
            }
        break;
        case 12:
          if ( d == 17L )
            {
              sprintf ( outbuf, "%s", "Holiday: National Day." );
              return (1);
            }
        break;

        default:
          ;
        break;
      }
    return (0);
  } // END - chk_west ()
  
int chk_sols ( long int t )  
  {
    long tx, G;
    long long S1, S2, S3, S4, S5, S6, G1, G2, G3, G4, G5, G6; // Sun values, weekday values
    int i, j, x;
    int more = 1;
     
//  printf ( "Starting solstice check\n" );
//  getch ();
    i = 0;
    for ( j = 0; j < 6; ++j )
      {
        s[i].tse[j] = tsebar[j];         
        s[i].nyi[j] = nyibar[j];         
      }

//    printf ( "\n1. %ld, Tsebar[%d]: ", t, i );
//    prn_lst ( s[i].tse );
//    printf ( "1. Nyibar[%d]: ", i );
//    prn_lst ( s[i].nyi );        
       
    i = 1;
    tx = t + 1L;
    do
      { 
//        printf ( "\nI = %d\n", i );
        tse_dru (tx);
        nyi_lon (tx);
        add_gen (tsebar, gzadru, tsedru, 7L, gza_f);
        add_gen (nyibar, nyidru, nyilon, 27L, sun_f );

        for ( j = 0; j < 6; ++j )
          {
            s[i].tse[j] = tsebar[j];         
            s[i].nyi[j] = nyibar[j];         
          }

//        printf ( "1. %ld, Tsebar[%d]: ", tx, i );
//        prn_lst ( s[i].tse );
//        printf ( "1. Nyibar[%d]: ", i );
//        prn_lst ( s[i].nyi );        

        if ( nyibar[1] > 45L ) 
           { 
             more = 0;
             x = i;
             for ( j = 0; j <= x; ++j )
               {
                 printf ( "\n2. Tsebar[%d]: ", j );
                 prn_lst ( s[j].tse );
                 printf ( "2. Nyibar[%d]: ", j );
                 prn_lst ( s[j].nyi );
               }
             sol_found = 1;                          
           }
          
        ++i;
        ++tx;
        if ( i > 7 )
          {
            printf ("Mega problem...!\n");
            more = 0;
            getch ();
          }
      } while ( more );
    
// The solstice is either in the solar day indicated by x or x-1.    

    G1 = (long long)(s[x-2].tse[0] * 60L * 60L * 6L * 707L); 
    G2 = (long long)(((( s[x-2].tse[0] * 60L + s[x-2].tse[1] ) * 60L 
         + s[x-2].tse[2] ) * 6L + s[x-2].tse[3] ) * 707L + s[x-2].tse[4]);
    G3 = (long long)(s[x-1].tse[0] * 60L * 60L * 6L * 707L); 
    G4 = (long long)(((( s[x-1].tse[0] * 60L + s[x-1].tse[1] ) * 60L 
         + s[x-1].tse[2] ) * 6L + s[x-1].tse[3] ) * 707L + s[x-1].tse[4]);         
    G5 = (long long)(s[x].tse[0] * 60L * 60L * 6L * 707L); 
    G6 = (long long)(((( s[x].tse[0] * 60L + s[x].tse[1] ) * 60L 
         + s[x].tse[2] ) * 6L + s[x].tse[3] ) * 707L + s[x].tse[4]);                  

    S2 = (long long)(((( s[x-2].nyi[0] * 60L + s[x-2].nyi[1] ) * 60L 
         + s[x-2].nyi[2] ) * 6L + s[x-2].nyi[3] ) * 67L + s[x-2].nyi[4]);
    S4 = (long long)(((( s[x-1].nyi[0] * 60L + s[x-1].nyi[1] ) * 60L 
         + s[x-1].nyi[2] ) * 6L + s[x-1].nyi[3] ) * 67L + s[x-1].nyi[4]);         
    S6 = (long long)(((( s[x].nyi[0] * 60L + s[x].nyi[1] ) * 60L 
         + s[x].nyi[2] ) * 6L + s[x].nyi[3] ) * 67L + s[x].nyi[4]);         

    S1 = S2 - (S4-S2) * (G2-G1)/(G4-G2);
    S3 = S2 + (S4-S2) * (G3-G2)/(G4-G2);    
    S5 = S4 + (S6-S4) * (G5-G4)/(G6-G4);        
    
//    printf ( "G1: %lld, G2: %lld, G3: %lld, G4: %lld, G5: %lld, G6: %lld\n", G1, G2, G3, G4, G5, G6 );
//    printf ( "S1: %lld, S2: %lld, S3: %lld, S4: %lld, S5: %lld, S6: %lld\n", S1, S2, S3, S4, S5, S6 );    

    if ( S3 > 27135000LL ) // x-2 day
      {
        sol_doweek = s[x-2].tse[0];
        sol1 = (long long) sol_doweek;
        G = ( 3600L * ( 27135000LL - S1 ) ) / (S3 - S1);
        sol2 = G / 60LL;
        sol3 = G % 60LL;
      }
    else if ( S3 <= 27135000LL && S5 > 27135000LL ) // X-1
      {
        sol_doweek = s[x-1].tse[0];
        sol1 = (long long) sol_doweek;
        G = ( 3600L * ( 27135000LL - S3 ) ) / (S5 - S3);
        sol2 = G / 60LL;
        sol3 = G % 60LL;
      }
    else         // X
      {
        sol_doweek = s[x].tse[0];    
        sol1 = (long long) sol_doweek;
        G = ( 3600LL *  (G6 - G5) ) / ( 60LL * 60LL * 6LL * 707LL );
        G = G * ( 27135000LL - S5 ) / (S6 - S5);
        sol2 = G / 60LL;
        sol3 = G % 60LL;
      }
    sol_doweek = sol_doweek + 1L;      
    if ( sol_doweek > 6L )
      sol_doweek = sol_doweek - 7L;
//    printf ( "DOWEEK FOUND: %ld\n", sol_doweek );
//    getch ();
  } // END - chk_sols

void set_lang ( void )
  {
    int i;
/******* Sanskrit *******
    for ( i = 0; i < 8; ++i )
      cycpar[i] = cycparE[i];
    for ( i = 0; i < 27; ++i )
      lunmanP[i] = lunmanS[i];
    for ( i = 0; i < 27; ++i )
      yogaP[i] = yogaS[i];
    for ( i = 0; i < 11; ++i )
      byedP[i] = byedS[i];
*******************/

/******* Tibetan ********/
    for ( i = 0; i < 27; ++i )
      lunmanP[i] = lunmanT[i];
    for ( i = 0; i < 27; ++i )
      yogaP[i] = yogaT[i];
    for ( i = 0; i < 11; ++i )
      byedP[i] = byedT[i];
/**************/
 } // END - set_lang
  
  void prn_bhut_cal ( void )
  {
    int i, more, n, newmth, lhag, chad, nextchad, nextlhag;
    int intercal, delaymth;
    int yr_gender;    // 1 = female, odd; 0 = male, even
    int mt_gender;    // 1 = female, odd; 0 = male, even
    long int mthnum; // Adjusted, without negative.
    long int cur_year, last_year;
    long int ltt, zla0, zla1;
    long int lastjd = 0L;
    long int curjd, nextjd; // Last and next JD calculations.
    int  lunmanx, sbyorba, byedpa, byedpax;
    int  prv_lunmanx, prv_sbyorba, prv_byedpa, prv_byedpax;
    long int Cmonth;
    long int sd_clunmanx;
    long int ld_parx; // sd_parx;
    long int ld_smex, sd_smex;
    long int yr_animx, mt_animx, ld_animx, sd_animx;
    long int yr_elemx, mt_elemx, /*ld_elemx,*/ sd_elemx;
    long int ZS_s, ZS_d, ZS_m;  // Zodiac signs, minutes, seconds;
    int e, g;
    int new_year;

    set_lang ();
    gza_short_flg = 0; // Flag for less accurate weekday adjustment
    cls ();
    pos (0,0);
    printf ( "%s\n", e_str[epch] );
    printf ("Enter the starting year: ");
    scanf ("%ld", &ty);
    printf ("\n");

    cur_year = ty;

    printf ("Enter the finish year: ");
    scanf ("%ld", &last_year );
    printf ("\n");

    donextyear:

    new_year = 1;
    ty = cur_year;
    tm = 1L;
    zeromthfg = 0;

    sprintf ( filenam, "bh_%ld.txt", ty );

    fptgt = fopen ( filenam, "w" );
    if ( fptgt == NULL )
      {
        printf ( "Error opening target file!\n" );
        getch ();
      }

// tt not yet set
// First, check to see if we are not at adjusted 12th month:

    zla_dag (ty, tm); // But only if moving one month at a time!

    zla0 = zladag[0];
    zla1 = zladag[1];

    adj_zla ();

    if ( adj_mth == 12L )
      {
        if ( !zeromthfg )       // Check this out again...
          ++tm;
      }
    else if ( adj_mth == -1L )
      {
        zladag[0] = zla0;
        zladag[1] = zla1;
        zeromthfg = 0;
      }
    n = 0;
    more = 1;
    newmth = 1;
    do   // START OF MONTHLY LOOP *******************************
      {
        intercal = 0;
        if ( adj_mth == -1L )
          intercal = 1;
        delaymth = 0;
        if ( newmth )
          {
            if ( !zeromthfg )   // We need to use the same data, twice.
              zla_dag (ty, tm); // But only if moving one month at a time!
            adj_zla ();
            cur_mth = zladag[0];  // NEEDED FOR GENERAL DAY
            gza_dru (cur_mth);
            nyi_dru (cur_mth);
            rilchaf ( cur_mth ); // MOVED, from gza_dag, NEW, 11/9/94
            // DO WE HAVE NEW YEAR?

            if ( ( adj_mth == -1 || adj_mth == 1 ) && new_year )
              {
                if ( ty < 7L )
                  {
                    yr_animx = ( ty + 1200L - 7L ) % 12L;
                    yr_elemx = ( ( ty + 1200L ) / 2L - 3L ) % 5L;
                    yr_gender = (int) ( ty + 1200L ) % 2; // 1 = female, odd; 0
                  }
                else
                  {
                    yr_animx = ( ty - 7L ) % 12L;
                    yr_elemx = ( ty / 2L - 3L ) % 5L;
                    yr_gender = (int) ty % 2; // 1 = female, odd; 0 = male
                  }
                sprintf ( outbuf, "\nNew Year: %ld, %s-%s-%s\n", ty,
                                  cycelem1[ (int) yr_elemx ],
                                  cycgendE[ yr_gender ],
                                  cycanim[ (int) yr_animx ] );
                fprintf ( fptgt, "%s", outbuf );
                new_year = 0;
              }
          }

// HAVE NOT CHECKED THAT THIS WORKS FOR TSHUR LUGS ???????

        if ( adj_mth < 0L )
          {
            intercal = 1;
            delaymth = 0;
          }
        else if ( adj_mth > 0L && ( zladag[1] == zlasho1 || zladag[1] == zlasho2 ) )
          {
            intercal = 0;
            delaymth = 1;
          }

        if ( adj_mth < 0L )
          mthnum = -adj_mth;
        else
          mthnum = adj_mth;

        printf ("The ZLA-DAG used is: %ld;%ld\n", zladag[0], zladag[1]);
//      cur_mth = zladag[0];    /**** Change other occurances ****/

        mt_animx = mthnum % 12L;
        Cmonth = mthnum + 2L;
        if ( Cmonth > 12L )
          Cmonth -= 12L;

        mt_elemx = yr_elemx + 1L + ( mthnum + 1L ) / 2L;

        if ( yr_gender == 0 ) // Male year, even CE number
          {
            if ( mthnum == 11L || mthnum == 12L )
              mt_elemx = yr_elemx + ( (int) mthnum + 1L ) / 2;
          }

        mt_elemx = mt_elemx % 5L;
        mt_gender = ( mthnum + 1 ) % 2; // There is no month zero

        if ( intercal )
          sprintf ( outbuf,
          "Bhutan Lunar Month: %ld (Intercalary) - %s-%s-%s",
          -adj_mth, cycelem1[ (int) mt_elemx ],
          cycgendE[ mt_gender ], cycanim[ (int) mt_animx ] );
        else
          sprintf ( outbuf, "Bhutan Lunar Month: %ld - %s-%s-%s",
          adj_mth, cycelem1[ (int) mt_elemx ],
          cycgendE[ mt_gender ], cycanim[ (int) mt_animx ] );

        printf ( "%s\n", outbuf );
        fprintf ( fptgt, "\n%s\n\n", outbuf );
        tt = 1L;
        ltt = 1L;
        do

// START OF DAY LOOP - START OF DAY LOOP - START OF DAY LOOP

          {   // START OF DAY LOOP *******************************
            tt = ltt;
            nextlhag = 0;
            lhag = 0;
            chad = 0;
            nextchad = 0;
//          printf ("Adjusted DATE is: %ld / %ld / %ld\n", tt, adj_mth, ty );
            tse_dru (tt);
            nyi_lon (tt);
            add_gen (tsebar, gzadru, tsedru, 7L, gza_f);
            add_gen (nyibar, nyidru, nyilon, 27L, sun_f );
            nyi_dag (nyibar); // MUST BE DONE FIRST
            gza_dag (tsebar);
            spi_zagf (); // calculates juldat

// Now, check for coming winter solstice:

            if ( nyibar[0] == 18L && nyibar[1] > 32L && nyibar[1] < 45L && sol_found == 0 ) 
              chk_sols ( tt );
            if ( nyibar[0] > 18L ) 
              sol_found = 0;   

// Work out previous and next days to get CHAD or LHAG
            curjd = juldat;
            tt = ltt - 1L;
            tse_dru (tt);
            nyi_lon (tt);
            add_gen (tsebar, gzadru, tsedru, 7L, gza_f);
            add_gen (nyibar, nyidru, nyilon, 27L, sun_f );
            nyi_dag (nyibar); // MUST BE DONE FIRST
            gza_dag (tsebar);
            spi_zagf ();
            lastjd = juldat;

            for ( i = 0; i < 6; ++i )
              {
                prv_nyibar[i] = nyibar[i];
                prv_nyidag[i] = nyidag[i];
                prv_tsebar[i] = tsebar[i];                
                prv_gzadag[i] = gzadag[i];
              }

// PROBABLY DON'T NEED THIS BLOCK - O YES WE DO

            tt = ltt + 1L;
            tse_dru (tt);
            nyi_lon (tt);
            add_gen (tsebar, gzadru, tsedru, 7L, gza_f);
            add_gen (nyibar, nyidru, nyilon, 27L, sun_f );
            nyi_dag (nyibar); // MUST BE DONE FIRST
            gza_dag (tsebar);
            spi_zagf ();
            nextjd = juldat;

            // Save for later, in case we have a LHAG day:

            for ( i = 0; i < 6; ++i )
              {
                nxt_nyibar[i] = nyibar[i];
                nxt_nyidag[i] = nyidag[i];
                nxt_tsebar[i] = tsebar[i];
                nxt_gzadag[i] = gzadag[i];
              }

// RECALCULATE FOR CURRENT TITHI:

            tt = ltt;
            tse_dru (tt);
            nyi_lon (tt);
            add_gen (tsebar, gzadru, tsedru, 7L, gza_f);
            add_gen (nyibar, nyidru, nyilon, 27L, sun_f );
            nyi_dag (nyibar); // MUST BE DONE FIRST
            gza_dag (tsebar);
            spi_zagf ();
            if ( curjd == lastjd + 2L )
              lhag = 1;
            if ( nextjd == curjd + 2L )
              nextlhag = 1;
            if ( curjd == lastjd )
              chad = 1;
            if ( curjd == nextjd )
              nextchad = 1;

//  CALCULATE LUNAR MANSION AT SUNRISE:

// add solar longitude to tithi * factor. Then subtract hours etc from gzadag

    clear_a_b ();
    for ( i = 0; i < 5; ++i )
      monlong[i] = 0L;

    monlong[1] = 54L;  // 1/30th of a revolution
    mul_gen ( monlong, monlong, tt, 27L, sun_f );
    add_gen ( monlong, monlong, nyidag, 27L, sun_f );
    listb[0] = 0L;
    for ( i = 1; i < 5; ++i )
      listb[i] = gzadag[i];
    listb[4] = ( sun_f * listb[4] ) / gza_f;
// This gives longitude at Daybreak:
    sub_gen ( monlong, monlong, listb, 27L, sun_f );

    lunmanx = (int) monlong[0];

// NOW CALCULATE SBYOR BA:

// This is strictly wrong, we need the sun's longitude at sunrise, but in the
// Tibetan tradition such an adjustment is not done

    add_gen ( lista, monlong, nyidag, 27L, sun_f );
    sbyorba = (int) lista[0];

    for ( i = 0; i < 6; ++i )
      sbyorlong[i] = lista[i];

// NOW CALCULATE BYED PA:

    clear_a_b ();
    sub_gen ( lista, monlong, nyidag, 27L, sun_f ); // AT SUNRISE FOR MOON
    byedpa = (int) lista[0] * 60 + (int) lista[1];
    byedpa = byedpa / 27;     // CHECK THIS
    if ( byedpa == 0 )
      byedpax = 7;
    else if ( byedpa == 57 )
      byedpax = 8;
    else if ( byedpa == 58 )
      byedpax = 9;
    else if ( byedpa == 59 )
      byedpax = 10;
    else
      byedpax = ( byedpa - 1 ) % 7;

//  printf ( "Julians: %ld, %ld, %ld\n", lastjd, curjd, nextjd );

// Solar day data:

            sd_animx = ( juldat - 2L ) % 12L;
            sd_elemx = ( ( juldat - 3L ) / 2L ) % 5L;
            sd_clunmanx = ( juldat - 17L ) % 28L;
            sd_smex = ( juldat - 2L ) % 9L + 1L;  // CHANGED 31/8/2004

// Lunar day data:

// If Chinese month is number 1, Trigram is Li, index = 1
//          ld_parx = ( mthnum - 1L ) * 30L + tt + 4L;
            ld_parx = ( Cmonth - 1L ) * 30L + tt;
            ld_parx = ld_parx % 8L;
// If Chinese month is number 1, "sme ba" is 1
//          ld_smex = ( mthnum - 1L ) * 30L + tt + 6L;
            ld_smex = ( Cmonth - 1L ) * 30L + tt;
            ld_smex = ld_smex % 9L;
            if ( ld_smex == 0L )
              ld_smex = 9L;
// If Chinese month is number 1, Animal is Tiger, index = 11
//          ld_animx = 6L * ( ( mthnum - 1L ) % 2L ) + tt + 10L;
            ld_animx = ( Cmonth - 1L ) * 30L + tt + 10L;
            ld_animx = ld_animx % 12L;

// FOR BHUTAN ONLY: !!!

            doweek = doweek + 1L; // doweek set by Julian routine
            if ( doweek == 7L )
              doweek = 0L;

            if ( chad )   // THIS IS FOR OMITTED LUNAR DAY
              {
                sprintf ( outbuf, "%ld. Omitted: %s %s %ld",
                          tt,
                          cycanim[ (int) ld_animx ],
                          cycpar[ (int) ld_parx ], ld_smex );
// Should include "gza'" for end of lunar day, plus solar longitude.
                fprintf ( fptgt, "%s\n", outbuf );
              }
            else  // IF NORMAL OR DUPLICATED:
              {
// NYIDAG IS NOT CHANGED AFTER HERE, SO WE CAN CALCULATE ZODIACAL SIGN:
// Zodiac signs, minutes, seconds -  ZS_s, ZS_d, ZS_m;

                ZS_m = ( ( nyidag[0] * 60L + nyidag[1] ) * 60L + nyidag[2] ) *
                         6L + nyidag[3];

// Changed this December 2006
                ZS_m = ( ( nyibar[0] * 60L + nyibar[1] ) * 60L + nyibar[2] ) *
                         6L + nyibar[3];

//              ZS_m = ( ZS_m * 21600L ) / 583200L;

                l2bcd ( bcdx0, ZS_m );
                mulbcdl ( bcdx0, bcdx0, 21600L );
                divbcdl ( bcdx1, bcdx0, 583200L );
                ZS_m = bcd2l ( bcdx1 );

                ZS_d = ZS_m / 60L;
                ZS_m = ZS_m % 60L;
                ZS_s = ZS_d / 30L;
                ZS_d = ZS_d % 30L;
                if ( lhag )  // IF THIS IS THE FIRST OF TWO WEEKDAYS.
                             // NEEDS TO BE REPLACED
                {
                // Put here any special calculations for extra day NEW
                // At this point, the gzadag, nyidag & monlong are correct for
                // the second of the two days with same tithi. We need for the
                // first.
                // Previous tithi ended just before start of day we need.
                clear_a_b ();
                for ( i = 0; i < 5; ++i )
                  prv_monlong[i] = 0L;
                prv_monlong[1] = 54L;

                mul_gen ( prv_monlong, prv_monlong, tt, 27L, sun_f );
                add_gen ( prv_monlong, prv_monlong, nyidag, 27L, sun_f );
                listb[0] = 1L;
                sub_gen ( prv_monlong, prv_monlong, listb, 27L, sun_f );

                prv_lunmanx = (int) prv_monlong[0];

                 // NOW CALCULATE SBYOR BA:

                  add_gen ( lista, prv_monlong, nyidag, 27L, sun_f );
                  prv_sbyorba = (int) lista[0];
                  for ( i = 0; i < 6; ++i )
                    prv_sbyorlong[i] = lista[i];

                // NOW CALCULATE SBYED PA:

                  clear_a_b ();
                  sub_gen ( lista, prv_monlong, prv_nyidag, 27L, sun_f );
                  prv_byedpa = (int) lista[0] * 60 + (int) lista[1];
                  prv_byedpa = prv_byedpa / 27; // CHECK THIS
                  if ( prv_byedpa == 0 )
                    prv_byedpax = 7;
                  else if ( prv_byedpa == 57 )
                    prv_byedpax = 8;
                  else if ( prv_byedpa == 58 )
                    prv_byedpax = 9;
                  else if ( prv_byedpa == 59 )
                    prv_byedpax = 10;
                  else
                    prv_byedpax = ( prv_byedpa - 1 ) % 7;

//              printf ( "Extra day\n" );
                juldat = juldat - 1L;

                jul2date ( juldat ); // This is normally done by spi_zagf

                doweek = doweek + 1L;
                 if ( doweek == 7L )
                   doweek = 0L;

// Solar day data:

            sd_animx = ( juldat - 2L ) % 12L;
            sd_elemx = ( ( juldat - 3L ) / 2L ) % 5L;
            sd_clunmanx = ( juldat - 17L ) % 28L;
            sd_smex = ( juldat - 2L ) % 9L + 1L;  // CHANGED 31/8/2004

// Lunar day data:

            ld_parx = ( mthnum - 1L ) * 30L + tt + 4L;
            ld_parx = ld_parx % 8L;
            ld_smex = ( mthnum - 1L ) * 30L + tt + 6L;
            ld_smex = ld_smex % 9L;
            if ( ld_smex == 0L )
              ld_smex = 9L;
            ld_animx = 6L * ( ( mthnum - 1L ) % 2L ) + tt + 10L;
            ld_animx = ld_animx % 12L;

            // LINE 1. First of duplicated
            sprintf ( outbuf, "%ld: %s. %s. %s-%s; %ld %s %ld",
                      tt,
                      dayoweek[ (int) doweek ],
                      lunmanP[ prv_lunmanx ],
                      fourelem[ dowkelem[ (int) doweek ] ],
                      fourelem[ lunmanelem[ prv_lunmanx ] ],
                      wd, wmonths[ (int) (wm-1L) ], wy);

//            printf ( "%s\n", outbuf );
              fprintf ( fptgt, "%s\n", outbuf );

            // LINE 2.
            sprintf ( outbuf, "%s, %s, %s, %s %ld",
                      yogaP[ prv_sbyorba ],
                      byedP[ prv_byedpax ],
                      cycanim[ ld_animx ],
                      cycpar[ (int) ld_parx ], ld_smex );
//              printf ( "%s\n", outbuf );
            fprintf ( fptgt, "\0x20\0x20%s\n", outbuf );

            // LINE 3.
            sprintf ( outbuf,
            "%ld;%ld,%ld %ld;%ld,%ld %ld;%ld,%ld %ld;%ld,%ld %ld;%ld,%ld",
                   ( prv_gzadag[0] + 1L ) % 7L, 60L, 0L,
//                     prv_gzadag[0], prv_gzadag[1], prv_gzadag[2],
                   prv_monlong[0], prv_monlong[1], prv_monlong[2],
                   nyidag[0], nyidag[1], nyidag[2],
                   prv_sbyorlong[0], prv_sbyorlong[1], prv_sbyorlong[2],
                   ZS_s, ZS_d, ZS_m );
            fprintf ( fptgt, "\x20\x20%s\n", outbuf );

            // LINE 4.
            sprintf ( outbuf, "Solar: %s-%s. %s %ld" ,
                      cycelem1[ (int) sd_elemx ],
                      cycanim[ (int) sd_animx ],
                      Clunman[ sd_clunmanx ],
                      sd_smex );
            fprintf ( fptgt, "\x20\x20%s\n", outbuf );
            // LINE 5. First of duplicated            
            if ( !delaymth && chk_spec ( mthnum, tt ) )
              {
//              printf ( "  %s\n", outbuf );
                fprintf ( fptgt, "    %s\n", outbuf );
              }       
            if ( !delaymth && chk_west ( wm, wd ) )
              {
//              printf ( "  %s\n", outbuf );
                fprintf ( fptgt, "    %s\n", outbuf );
              }                          
            if ( sol_found && sol_doweek == doweek )
              {
                sprintf ( outbuf, "Winter solstice, time: %lld;%lld,%lld", sol1, sol2, sol3 );        
//              printf ( "  %s\n", outbuf );
                fprintf ( fptgt, "    %s\n", outbuf );
              }              

      // NOW DO SECOND OF REPEATED LUNAR DAYS
      // NEEDS TO BE REPLACED
            juldat = juldat + 1L;  // THIS IS READJUSTMENT!
            jul2date ( juldat );
            doweek = doweek + 1L;
            if ( doweek == 7L )
              doweek = 0L;
// Solar day data:   // This is a brute force method I developed.

            sd_animx = ( juldat - 2L ) % 12L;
            sd_elemx = ( ( juldat - 3L ) / 2L ) % 5L;
            sd_clunmanx = ( juldat - 17L ) % 28L;
            sd_smex = ( juldat - 2L ) % 9L + 1L;  // CHANGED 31/8/2004

                // LINE 1. Second of duplicated.
            sprintf ( outbuf, "%ld: %s. %s. %s-%s; %ld %s %ld",
                      tt,
                      dayoweek[ (int) doweek ],
                      lunmanP[ lunmanx ],
                      fourelem[ dowkelem[ (int) doweek ] ],
                      fourelem[ lunmanelem[ lunmanx ] ],
                      wd, wmonths[ (int) (wm-1L) ], wy);
            fprintf ( fptgt, "%s\n", outbuf );

            // LINE 2.
            sprintf ( outbuf, "%s, %s, %s, %s %ld",
                      yogaP[ sbyorba ],
                      byedP[ byedpax ],
                      cycanim[ ld_animx ],
                      cycpar[ (int) ld_parx ], ld_smex );
//              printf ( "%s\n", outbuf );
            fprintf ( fptgt, "\x20\x20%s\n", outbuf );
            // LINE 3.
            sprintf ( outbuf,
            "%ld;%ld,%ld %ld;%ld,%ld %ld;%ld,%ld %ld;%ld,%ld %ld;%ld,%ld",
                      gzadag[0], gzadag[1], gzadag[2],
                      monlong[0], monlong[1], monlong[2],
                      nyidag[0], nyidag[1], nyidag[2],
                      sbyorlong[0], sbyorlong[1], sbyorlong[2],
                      ZS_s, ZS_d, ZS_m );
            fprintf ( fptgt, "\x20\x20%s\n", outbuf );
            // LINE 4.
            sprintf ( outbuf, "Solar: %s-%s. %s %ld" ,
                      cycelem1[ (int) sd_elemx ],
                      cycanim[ (int) sd_animx ],
                      Clunman[ sd_clunmanx ],
                      sd_smex );
            fprintf ( fptgt, "\x20\x20%s\n", outbuf );
            // LINE 5. Second of duplicated
            if ( !delaymth && chk_west ( wm, wd ) )
              {
//              printf ( "  %s\n", outbuf );
                fprintf ( fptgt, "    %s\n", outbuf );
              }   
            if ( sol_found && sol_doweek == doweek )
              {
                sprintf ( outbuf, "Winter solstice, time: %lld;%lld,%lld", sol1, sol2, sol3 );        
//              printf ( "  %s\n", outbuf );
                fprintf ( fptgt, "    %s\n", outbuf );
              }                                                                   
            }
          else  // NO LHAG - NORMAL DAY - REPLACED
            {
                // LINE 1. Normal day
                sprintf ( outbuf, "%ld: %s. %s. %s-%s; %ld %s %ld",
                          tt,
                          dayoweek[ (int) doweek ],
                          lunmanP[ lunmanx ],
                          fourelem[ dowkelem[ (int) doweek ] ],
                          fourelem[ lunmanelem[ lunmanx ] ],
                          wd, wmonths[ (int) (wm-1L) ], wy);

//              printf ( "%s\n", outbuf );
                fprintf ( fptgt, "%s\n", outbuf );

                // LINE 2.
                sprintf ( outbuf, "%s, %s, %s, %s %ld",
                          yogaP[ sbyorba ],
                          byedP[ byedpax ],
                          cycanim[ ld_animx ],
                          cycpar[ (int) ld_parx ], ld_smex );
//              printf ( "%s\n", outbuf );
                fprintf ( fptgt, "\x20\x20%s\n", outbuf );

                // LINE 3.
                sprintf ( outbuf,
                "%ld;%ld,%ld %ld;%ld,%ld %ld;%ld,%ld %ld;%ld,%ld %ld;%ld,%ld",
                          gzadag[0], gzadag[1], gzadag[2],
                          monlong[0], monlong[1], monlong[2],
                          nyidag[0], nyidag[1], nyidag[2],
                          sbyorlong[0], sbyorlong[1], sbyorlong[2],
                          ZS_s, ZS_d, ZS_m );
                fprintf ( fptgt, "\x20\x20%s\n", outbuf );

                // LINE 4.
                sprintf ( outbuf, "Solar: %s-%s. %s %ld" ,
                          cycelem1[ (int) sd_elemx ],
                          cycanim[ (int) sd_animx ],
                          Clunman[ sd_clunmanx ],
                          sd_smex );
                fprintf ( fptgt, "\x20\x20%s\n", outbuf );
                
                if ( !delaymth && chk_spec ( mthnum, tt ) )
                  {
//                  printf ( "  %s\n", outbuf );
                    fprintf ( fptgt, "    %s\n", outbuf );
                  }                
                if ( !delaymth && chk_west ( wm, wd ) )
                  {
//                  printf ( "  %s\n", outbuf );
                    fprintf ( fptgt, "    %s\n", outbuf );
                  }
                if ( sol_found && sol_doweek == doweek )
                  {
                    sprintf ( outbuf, "Winter solstice, time: %lld;%lld,%lld", sol1, sol2, sol3 );        
//                  printf ( "  %s\n", outbuf );
                    fprintf ( fptgt, "    %s\n", outbuf );
                  }                                
                } // END - NO LHAG - NORMAL DAY
              }

            ++ltt;
            newmth = 0;
            lastjd = juldat;
          } while ( ltt <= 30 ); // END OF DAY LOOP

// END OF DAY LOOP - END OF DAY LOOP - END OF DAY LOOP - END OF DAY LOOP -

        newmth = 1;
        if ( !zeromthfg )
          ++tm;
        if ( tm > 12 )
          {
            ++ty;
            tm = 1;
          }
        if ( adj_mth == 12L )
          more = 0; // We have finished.
      } while ( more );        // END OF MONTH LOOP
    fclose ( fptgt );

    if ( cur_year < last_year )
      {
        ++cur_year;
        goto donextyear;
      }

  } // END - prn_bhut_cal ()
